package pl.asie.ucw;

import com.google.common.collect.ImmutableMap;
import net.minecraft.block.state.IBlockState;
import net.minecraft.client.renderer.block.model.*;
import net.minecraft.client.renderer.texture.TextureAtlasSprite;
import net.minecraft.client.renderer.vertex.VertexFormat;
import net.minecraft.entity.EntityLivingBase;
import net.minecraft.item.ItemStack;
import net.minecraft.util.EnumFacing;
import net.minecraft.util.ResourceLocation;
import net.minecraft.world.World;
import net.minecraftforge.client.model.IModel;
import net.minecraftforge.client.model.pipeline.UnpackedBakedQuad;
import net.minecraftforge.common.model.IModelState;
import org.apache.commons.lang3.tuple.Pair;

import javax.annotation.Nullable;
import javax.vecmath.Matrix4f;
import java.util.*;
import java.util.function.Function;

public class TintApplyingModel implements IModel {
	public static class IOL extends ItemOverrideList {
		private final ItemOverrideList parent;

		public IOL(ItemOverrideList parent) {
			super(Collections.emptyList());
			this.parent = parent;
		}

		@Override
		public ResourceLocation func_188021_a(ItemStack stack, @Nullable World worldIn, @Nullable EntityLivingBase entityIn) {
			return parent.func_188021_a(stack, worldIn, entityIn);
		}

		@Override
		public com.google.common.collect.ImmutableList<ItemOverride> getOverrides() {
			return parent.getOverrides();
		}

		@Override
		public IBakedModel handleItemState(IBakedModel originalModel, ItemStack stack, @Nullable World world, @Nullable EntityLivingBase entity) {
			IBakedModel trueParent = originalModel;
			if (originalModel instanceof Baked) {
				trueParent = ((Baked) originalModel).parent;
			}
			IBakedModel model = parent.handleItemState(trueParent, stack, world, entity);
			if (model == trueParent) {
				return originalModel;
			} else {
				return new Baked(model);
			}
		}
	}

	public static class Baked implements IBakedModel {
		private final IBakedModel parent;

		public Baked(IBakedModel parent) {
			this.parent = parent;
		}

		@Override
		public List<BakedQuad> func_188616_a(@Nullable IBlockState state, @Nullable EnumFacing side, long rand) {
			List<BakedQuad> orig = parent.func_188616_a(state, side, rand);
			if (orig != null && !orig.isEmpty()) {
				List<BakedQuad> newList = new ArrayList<>(orig.size());
				for (BakedQuad q : orig) {
					if (!q.func_178212_b()) {
						// TODO: Handle UBQs better
						newList.add(new BakedQuad(
								q.func_178209_a(),
								0,
								q.func_178210_d(),
								q.func_187508_a(),
								q.shouldApplyDiffuseLighting(),
								q.getFormat()
						));
					}
				}
				return newList;
			} else {
				return orig;
			}
		}

		@Override
		public boolean func_177555_b() {
			return parent.func_177555_b();
		}

		@Override
		public boolean func_177556_c() {
			return parent.func_177556_c();
		}

		@Override
		public boolean func_188618_c() {
			return parent.func_188618_c();
		}

		@Override
		public TextureAtlasSprite func_177554_e() {
			return parent.func_177554_e();
		}

		@Override
		public ItemOverrideList func_188617_f() {
			return new IOL(parent.func_188617_f());
		}

		@Override
		public ItemCameraTransforms func_177552_f() {
			return parent.func_177552_f();
		}

		@Override
		public org.apache.commons.lang3.tuple.Pair<? extends IBakedModel, javax.vecmath.Matrix4f> handlePerspective(ItemCameraTransforms.TransformType cameraTransformType) {
			Pair<? extends IBakedModel, Matrix4f> pair = parent.handlePerspective(cameraTransformType);
			if (pair.getLeft() == parent) {
				return Pair.of(this, pair.getRight());
			} else {
				return Pair.of(new Baked(pair.getLeft()), pair.getRight());
			}
		}
	}

	private final IModel parent;

	public TintApplyingModel(IModel parent) {
		this.parent = parent;
	}

	@Override
	public Collection<ResourceLocation> getDependencies() {
		return parent.getDependencies();
	}

	@Override
	public Collection<ResourceLocation> getTextures() {
		return parent.getTextures();
	}

	@Override
	public IModelState getDefaultState() {
		return parent.getDefaultState();
	}

	@Override
	public IBakedModel bake(IModelState state, VertexFormat format, Function<ResourceLocation, TextureAtlasSprite> bakedTextureGetter) {
		return new Baked(parent.bake(state, format, bakedTextureGetter));
	}

	private IModel wrap(IModel other) {
		if (other == parent) {
			return this;
		} else {
			return new TintApplyingModel(other);
		}
	}

	@Override
	public IModel process(ImmutableMap<String, String> customData) {
		return wrap(parent.process(customData));
	}

	@Override
	public IModel smoothLighting(boolean value) {
		return wrap(parent.smoothLighting(value));
	}

	@Override
	public IModel gui3d(boolean value) {
		return wrap(parent.gui3d(value));
	}

	@Override
	public IModel uvlock(boolean value) {
		return wrap(parent.uvlock(value));
	}

	@Override
	public IModel retexture(ImmutableMap<String, String> textures) {
		return wrap(parent.retexture(textures));
	}
}
